# CheapEmote by David Bowland
# ./addons/eventscripts/cheapemote/cheapemote.py

# >>> To configure this addon please see cheapemote.cfg (created when the addon is first loaded) <<<

"""
Provides chat triggers to display emote icons above player heads.
Requires only ES 2.0.0.250c+
"""

import cfglib
import cmdlib
import es
import gamethread
import os.path
import re
import time

import psyco
psyco.full()


info = es.AddonInfo()
info.name          = 'CheapEmote'
info.version       = '3'
info.basename      = 'cheapemote'
info.url           = 'http://addons.eventscripts.com/addons/view/' + info.basename
info.author        = 'SuperDave'
info.specialthanks = 'Einlanzers, Mattie, Venjax, and Wonder'

###

config = cfglib.AddonCFG(os.path.join(es.getAddonPath(info.basename), info.basename + '.cfg'))

config.text(info.name + ' release %s options' % info.version)
config.text('./addons/eventscripts/%(basename)s/%(basename)s.cfg' % {'basename': info.basename})
config.text(info.url)
config.text('')
config.text('Load this addon with: es_load ' + info.basename)
config.text('')
config.text('Thanks to Einlanzers, Mattie, Venjax, and Wonder for their work on the original script!')
config.text('')
config.text('All models included with CheapEmote are by Venjax. The original Emote can be found here:')
config.text('http://addons.eventscripts.com/addons/view/emote')
config.text('\n')

config.text('***** Delay option *****')
icon_delay = config.cvar('cheapemote_expiredelay', 5, 'Number of seconds to show emote icon')
config.text('\n')

config.text('***** Sound options *****')
sound_delay  = config.cvar('cheapemote_sound_delay',       5, 'Number of seconds after triggering a sound a player must wait to trigger another sound')
sound_volume = config.cvar('cheapemote_sound_volume',      1, 'Volume to play sounds (0.0 = no sound, 1.0 = full volume)')
attenuation  = config.cvar('cheapemote_sound_attenuation', 0, 'Attenuation to use when playing sounds (0.0 = entire map, 1.0 = single player)')
config.text('\n')

config.text('***** Icon/trigger options *****')
config.text('')
config.text('Use the following server command to add emote icons/triggers:')
config.text('')
config.text('cheapemote_addicon <"trigger"> <"icon path"> ["sound relative to ./sound/"]')
config.text('')
config.text('Each icon requires:')
config.text('   ./materials/models/extras/<icon path>/speech_info.vmt')
config.text('   ./materials/models/extras/<icon path>/speech_info.vtf')
config.text('   ./models/extras/<icon path>/info_speech.dx80.vtx')
config.text('   ./models/extras/<icon path>/info_speech.dx90.vtx')
config.text('   ./models/extras/<icon path>/info_speech.mdl')
config.text('   ./models/extras/<icon path>/info_speech.phy')
config.text('   ./models/extras/<icon path>/info_speech.sw.vtx')
config.text('   ./models/extras/<icon path>/info_speech.vvd')
config.text('\n')

config.text('Place your cheapemote_addicon commands below:')
config.command('cheapemote_addicon')
config.text('\n')

config.text('Uncomment the following default options if you wish to use them:')
config.text('')
config.text('cheapemote_addicon "m!" "mbang"')
config.text('cheapemote_addicon ":)" "smile"')
config.text('cheapemote_addicon ":(" "frown"')
config.text('cheapemote_addicon "lol" "grin"')
config.text('cheapemote_addicon ":o" "ooh"')
config.text('cheapemote_addicon ":P" "ras"')
config.text('\n')

config.text('The following server command can be used to remove emote triggers:')
config.text('')
config.text('cheapemote_removeicon <"trigger">')
config.text('')
config.text('Due to the fact emote icons/triggers can be added and removed above, this command will not be used by most users.')

config.write()


gamename = es.getGameName()

###

class EmoteIcon(object):
   lastgive = es.ServerVar('eventscripts_lastgive') # Provided by ES for the last created entity index

   index = 0
   lastsound = 0

   def __init__(self, userid):
      self.userid = userid

   def createIcon(self, icon, sound):
      """ Creates the specified icon above the player """
      self.removeIcon()

      # Play a sound if the sound doesn't follow the last sound too quickly
      now = time.time()
      if sound and now - self.lastsound >= sound_delay:
         es.emitsound('player', self.userid, sound, sound_volume, attenuation)
         self.lastsound = now

      # Remember eye angles
      eye_angles = tuple(es.getplayerprop(self.userid, players.eyeangle_prop % x) for x in (0, 1))
      # Make the player look straight up
      es.server.queuecmd('es_xsetang %s 90 0 0' % self.userid)

      # Create the icon (dynamic prop), use a server command to remember the entity index of the icon,
      # and set the name of the icon entity to a unique name so it can be individually manipulated
      es.server.queuecmd('es_xprop_dynamic_create %(userid)s "extras/%(icon)s/info_speech.mdl";cheapemote_setindex %(userid)s;es_xentsetname %(userid)s emote_bubble_%(userid)s' % {'userid': self.userid, 'icon': icon})
      # Grab the player's location and set the icon's origin to 96 units above the player's feet
      location = es.getplayerlocation(self.userid)
      es.server.queuecmd('es_xfire %s emote_bubble_%s addoutput "origin %s %s %s"' % (self.userid, self.userid, location[0], location[1], location[2] + 96))
      # Set the icon to be parented to the player so it follows the player around
      es.server.queuecmd('es_xfire %(userid)s emote_bubble_%(userid)s SetParent !activator;es_xfire %(userid)s emote_bubble_%(userid)s SetAnimation idle;es_xfire %(userid)s emote_bubble_%(userid)s addoutput "angles 0 0 0";es_xfire %(userid)s emote_bubble_%(userid)s addoutput "solid 0"' % {'userid': self.userid})

      # Revert the player's eye angles to their original values
      es.server.queuecmd('es_xsetang %s %s %s 0' % ((self.userid,) + eye_angles))

      # Start the delay to remove the icon
      gamethread.delayedname(icon_delay, self.getDelayName(self.userid), self.removeIcon, False)

   def removeIcon(self, remove_delay=True):
      """ Removes the player's icon """
      if self.index:
         temp_userid = es.getuserid()
         if temp_userid:
            es.server.queuecmd('es_xfire %s emote_bubble_%s kill' % (temp_userid, self.userid))
         else:
            # If the stored index is still a prop_dynamic entity we remove it
            entlist = es.createentitylist(self.index)
            if (entlist[self.index]['classname'] == 'prop_dynamic') if entlist else False:
               es.server.queuecmd('es_xremove %s' % self.index)

         # We also cancel the remove delay and reset the icon index to zero
         if remove_delay:
            self.removeDelay()
         self.index = 0

   def removeDelay(self):
      """ Removes any outstanding icon delay """
      if self.index:
         gamethread.cancelDelayed(self.getDelayName(self.userid))

   def setIndex(self):
      """ Retrieves the icon's entity index """
      self.index = int(self.lastgive)

   @staticmethod
   def getDelayName(userid):
      """ Returns the unique name for the player's emote icon """
      return 'emote_' + str(userid)

###

class EmotePlayers(dict):
   """ Contains players with emote icons """
   eye_angles = {'cstrike': 'CCSPlayer.m_angEyeAngles[%s]',
                 'dod': 'CDODPlayer.m_angEyeAngles[%s]',
                 'hl2mp': 'CHL2MP_Player.m_angEyeAngles[%s]',
                 'tf': 'CTFPlayer.tfnonlocaldata.m_angEyeAngles[%s]',
                 'left4dead': 'CCSPlayer.m_angEyeAngles[%s]'}

   def __init__(self):
      if gamename not in self.eye_angles:
         raise NotImplementedError, "CheapEmote does not support game '%s'" % gamename
      self.eyeangle_prop = self.eye_angles[gamename]
      cmdlib.registerServerCommand('cheapemote_setindex', self.setIndexCmd, 'CheapEmote: INTERNAL USE ONLY')

      super(EmotePlayers, self).__init__()

   """ Dictionary """

   def __getitem__(self, userid):
      """ Returns the EmoteIcon instance corresponding to the userid """
      userid = int(userid)
      if userid not in self:
         self[userid] = EmoteIcon(userid)
      return super(EmotePlayers, self).__getitem__(userid)

   def __delitem__(self, userid):
      """ Removes the EmoteIcon instance corresponding to the userid """
      userid = int(userid)
      if userid in self:
         self[userid].removeIcon()
         super(EmotePlayers, self).__delitem__(userid)

   def clear(self, remove=False):
      """ Clears all icons """
      for userid in self:
         if remove:
            self[userid].removeIcon()
         else:
            self[userid].removeDelay()
      super(EmotePlayers, self).clear()

   """ Unique functions """

   def setIndexCmd(self, args):
      """ Sets the entity index of the player's icon """
      userid = int(args[0])
      if userid in self:
         self[userid].setIndex()

   @staticmethod
   def unregisterCommands():
      """ Unregisters player-based server commands """
      cmdlib.unregisterServerCommand('cheapemote_setindex')

players = EmotePlayers()

###

class IconManager(object):
   soundpath = os.path.join(str(es.ServerVar('eventscripts_gamedir')), 'sound', '')

   def __init__(self):
      self.icons = {}

      cmdlib.registerServerCommand('cheapemote_addicon', self.addIconCmd,
       'cheapemote_addicon <"trigger"> <"icon path">  ["sound relative to ./sound/"]\nAdds an emote icon')
      cmdlib.registerServerCommand('cheapemote_removeicon', self.removeIconCmd, 'cheapemote_removeicon <"trigger">\nRemoves an emote icon')

   """ Icon management functions """

   def addIcon(self, trigger, icon, sound=''):
      """ Adds the icon and trigger to the dictionary of icons """
      self.icons[trigger] = {'icon': icon, 'regex': re.compile('([ \\.!\\@\\$%\\^\\&\\*\\(\\),\";\\?-]|^)' + re.escape(trigger) +
       '([ \\.!\\@\\$%\\^\\&\\*\\(\\),\";\\?-]|$)'), 'sound': sound}

   def removeIcon(self, trigger):
      """ Removes the trigger from the dictionary of icons """
      if trigger in self.icons:
         del self.icons[trigger]

   def findIcon(self, userid, text):
      """ Displays any icons corresponding with the text """
      if not es.getplayerprop(userid, 'CBasePlayer.pl.deadflag'):
         for trigger in self.icons:
            current_icon = self.icons[trigger]
            if current_icon['regex'].search(text):
               players[userid].createIcon(current_icon['icon'], current_icon['sound'])
               break

   def markDownloads(self):
      """ Marks icon materials for download """
      for trigger in self.icons:
         current_icon = self.icons[trigger]
         for path in ('materials/models/extras/%s/speech_info.vmt', 'materials/models/extras/%s/speech_info.vtf',
          'models/extras/%s/info_speech.dx80.vtx', 'models/extras/%s/info_speech.dx90.vtx', 'models/extras/%s/info_speech.mdl',
          'models/extras/%s/info_speech.phy', 'models/extras/%s/info_speech.sw.vtx', 'models/extras/%s/info_speech.vvd'):
            es.stringtable('downloadables', path % current_icon['icon'])
         if os.path.isfile(self.soundpath + current_icon['sound']) if current_icon['sound'] else False:
            es.stringtable('downloadables', 'sound/' + current_icon['sound'])

   """ Server command functions """

   def addIconCmd(self, args):
      """
      cheapemote_addicon <"trigger"> <"icon path"> ["sound relative to ./sound/"]
      """
      if len(args) in (2, 3):
         if len(args) == 2:
            self.addIcon(args[0], args[1])
         else:
            self.addIcon(args[0], args[1], args[2])
      else:
         es.dbgmsg(0, 'Syntax: cheapemote_addicon <"trigger"> <"icon path"> ["sound relative to ./sound/"]')

   def removeIconCmd(self, args):
      """
      cheapemote_removeicon <"trigger">
      """
      if len(args) == 1:
         icons.removeIcon(args[0])
      else:
         es.dbgmsg(0, 'Syntax: cheapemote_removeicon <"trigger">')

   """ Static functions """

   @staticmethod
   def unregisterCommands():
      """ Unregisters icon-based server commands """
      cmdlib.unregisterServerCommand('cheapemote_addicon')
      cmdlib.unregisterServerCommand('cheapemote_removeicon')

icons = IconManager()

###

def load():
   config.execute()
   if str(es.ServerVar('eventscripts_currentmap')):
      icons.markDownloads()


def es_map_start(event_var):
   players.clear()
   icons.markDownloads()


def round_start(event_var):
   players.clear()


def dod_round_start(event_var):
   players.clear()


def player_say(event_var):
   icons.findIcon(int(event_var['userid']), event_var['text'])


def player_death(event_var):
   del players[event_var['userid']]


def player_disconnect(event_var):
   del players[event_var['userid']]


def unload():
   players.clear(True)

   icons.unregisterCommands()
   players.unregisterCommands()

   for cvar in config.getCvars():
      es.ServerVar(cvar).set(0)